@@ -14,9 +14,12 @@ sub many_to_many {
     my $remove_meth = "remove_from_${meth}";
     my $set_meth = "set_${meth}";
 
+    $rel_attrs->{alias} ||= $f_rel;
+
     *{"${class}::${meth}"} = sub {
       my $self = shift;
       my $attrs = @_ > 1 && ref $_[$#_] eq 'HASH' ? pop(@_) : {};
+      my @args = ($f_rel, @_ > 0 ? @_ : undef, { %{$rel_attrs||{}}, %$attrs });
       $self->search_related($rel)->search_related(
         $f_rel, @_ > 0 ? @_ : undef, { %{$rel_attrs||{}}, %$attrs }
       );
@@ -32,10 +35,19 @@ sub many_to_many {
       my $rel_source_name = $source->relationship_info($rel)->{source};
       my $rel_source = $schema->resultset($rel_source_name)->result_source;
       my $f_rel_source_name = $rel_source->relationship_info($f_rel)->{source};
-      my $f_rel_rs = $schema->resultset($f_rel_source_name);
-      my $obj = ref $_[0]
-        ? ( ref $_[0] eq 'HASH' ? $f_rel_rs->create($_[0]) : $_[0] )
-        : ( $f_rel_rs->create({@_}) );
+      my $f_rel_rs = $schema->resultset($f_rel_source_name)->search({}, $rel_attrs||{});
+
+      my $obj;
+      if (ref $_[0]) {
+        if (ref $_[0] eq 'HASH') {
+          $obj = $f_rel_rs->create($_[0]);
+        } else {
+          $obj = $_[0];
+        }
+      } else {
+        $obj = $f_rel_rs->create({@_});
+      }
+
       my $link_vals = @_ > 1 && ref $_[$#_] eq 'HASH' ? pop(@_) : {};
       my $link = $self->search_related($rel)->new_result({});
       $link->set_from_related($f_rel, $obj);
@@ -48,8 +60,9 @@ sub many_to_many {
       @_ > 0 or $self->throw_exception(
         "{$set_meth} needs a list of objects or hashrefs"
       );
+      my @to_set = (ref($_[0]) eq 'ARRAY' ? @{ $_[0] } : @_);
       $self->search_related($rel, {})->delete;
-      $self->$add_meth(shift) while (defined $_[0]);
+      $self->$add_meth($_) for (@to_set);
     };
 
     *{"${class}::${remove_meth}"} = sub {
