@@ -8,7 +8,9 @@ use base qw/DBIx::Class/;
 
 sub iterator_class  { shift->result_source_instance->resultset_class(@_) }
 sub resultset_class { shift->result_source_instance->resultset_class(@_) }
+sub result_class { shift->result_source_instance->result_class(@_) }
 sub source_name { shift->result_source_instance->source_name(@_) }
+sub source_info { shift->result_source_instance->source_info(@_) }
 
 sub resultset_attributes {
   shift->result_source_instance->resultset_attributes(@_);
@@ -26,22 +28,23 @@ sub add_columns {
 *add_column = \&add_columns;
 
 sub has_column {
-  my ($self, $column) = @_;
-  return $self->result_source_instance->has_column($column);
+  shift->result_source_instance->has_column(@_);
 }
 
 sub column_info {
-  my ($self, $column) = @_;
-  return $self->result_source_instance->column_info($column);
+  shift->result_source_instance->column_info(@_);
 }
 
+sub column_info_from_storage {
+  shift->result_source_instance->column_info_from_storage(@_);
+}
 
 sub columns {
-  return shift->result_source_instance->columns(@_);
+  shift->result_source_instance->columns(@_);
 }
 
 sub remove_columns {
-  return shift->result_source_instance->remove_columns(@_);
+  shift->result_source_instance->remove_columns(@_);
 }
 
 *remove_column = \&remove_columns;
