@@ -3,22 +3,23 @@ package DBIx::Class::Storage::DBI::SQLite;
 use strict;
 use warnings;
 
-use base qw/DBIx::Class::Storage::DBI/;
+use base qw/DBIx::Class::Storage::DBI::MultiDistinctEmulation/;
 
-sub last_insert_id {
-  return $_[0]->dbh->func('last_insert_rowid');
+sub _dbh_last_insert_id {
+  my ($self, $dbh, $source, $col) = @_;
+  $dbh->func('last_insert_rowid');
 }
 
 1;
 
-=head1 NAME 
+=head1 NAME
 
 DBIx::Class::PK::Auto::SQLite - Automatic primary key class for SQLite
 
 =head1 SYNOPSIS
 
   # In your table classes
-  __PACKAGE__->load_components(qw/PK::Auto::SQLite Core/);
+  __PACKAGE__->load_components(qw/PK::Auto Core/);
   __PACKAGE__->set_primary_key('id');
 
 =head1 DESCRIPTION
